module elas2d_distrib_type_mod
	integer,parameter :: P_i=1, S_i=2
	integer,parameter :: ux_i=3, uz_i=4
	integer,parameter :: vx_i=5, vz_i=6
	integer,parameter :: txx_i=7, tzz_i=8, txz_i=9
	integer,parameter :: Upx_i=10, Upz_i=11, Usx_i=12, Usz_i=13
	integer,parameter :: uzdz_i=14


type distrib_type
	integer :: hx,lx,hz,lz ! thread index for my neighbors
	integer :: nx,nz          ! my size
	integer :: n_x,n_z          ! my size unhaloed
	real,pointer :: dx(:),dz(:)  ! my sampling in dx and dz
	real         :: dt        ! my sampling in dt
	integer      :: fx,fz     ! my shift in x, and z
	integer      :: f_x,f_z     ! my shift in x, and z unhaloed
	real,pointer :: rho(:,:)   ! smooth density
	real,pointer :: lambda(:,:)! smooth lambda
	real,pointer :: mu(:,:)    ! smooth mu
	real,pointer :: P (:,:)    ! Pressure wave
	real,pointer :: S (:,:)    ! Shear wave
	real,pointer :: ux(:,:)    ! ux
	real,pointer :: uz(:,:)    ! uz
	real,pointer :: vx(:,:)    ! vx
	real,pointer :: vz(:,:)    ! vz
	real,pointer :: txx(:,:)   ! txx
	real,pointer :: tzz(:,:)   ! tzz
	real,pointer :: txz(:,:)   ! txz
	real,pointer :: Upx(:,:)
	real,pointer :: Upz(:,:)
	real,pointer :: Usx(:,:)
	real,pointer :: Usz(:,:)
	real,pointer :: uzdz(:,:)
	real,pointer :: sponge_x(:)  ! sponge in x
	real,pointer :: sponge_z(:)  ! sponge in z
	real,pointer :: cx1(:,:),cx2(:,:),cz1(:,:),cz2(:,:)
	integer,pointer ::recs (:,:)  ! receiver positions
	real,pointer :: vx_data (:,:) ! receiver data
	real,pointer :: vz_data (:,:)
	real,pointer :: P_data  (:,:)
	real,pointer  ::  all(:,:,:) !all of the arrays
	integer           :: ixsrc,izsrc
	real, pointer     :: buf1(:),buf2(:)
end type

end module


module elas2d_distrib_mod
	use elas2d_distrib_type_mod
	use elas2d_thread_type_mod
	use elas2d_model_type_mod
	use elas2d_helixcartmod
	use sep
	implicit none
	real, private, pointer, save :: b_l_x(:,:),b_u_x(:,:)
	real, private, pointer, save :: b_l_z(:,:),b_u_z(:,:)
	integer,external,private :: omp_get_thread_num


	contains


	subroutine distrib_init (thread_s,model_s,distrib_s)
		type(thread_type)			:: thread_s
		type(model_type)			:: model_s
		type(distrib_type),pointer	:: distrib_s(:)

		integer						:: npx, npz, n
		integer						:: iz, ix, ic
		integer						:: bz, ez, bx, ex, d, dx, dz, imax
		integer						:: n_x, n_y, n_z, irec, nr, ii
		integer						:: fx, fz, ith, i(2), nrec_found

		integer,allocatable			:: nnx(:), nnz(:),rr(:,:)
		integer,allocatable			:: nfx(:), nfz(:)
		integer,allocatable			:: n_nx(:), n_nz(:)
		integer,allocatable			:: n_fx(:), n_fz(:)

		real						:: a0 
		

		call from_param ("npx", npx, 1)
		call from_param ("npz", npz, 1)

		!allocate the boundaries
		n_x = ceiling (real (model_s%nx) / real (npx))
		n_z = ceiling (real (model_s%nz) / real (npz))

		allocate (b_l_x (n_x+2,thread_s%nth), b_u_x (n_x+2,thread_s%nth))
		allocate (b_l_z (n_z+2,thread_s%nth), b_u_z (n_z+2,thread_s%nth))

		allocate (n_nx (npx), n_fx (npx))
		allocate (n_nz (npz), n_fz (npz))

		allocate (nnx (npx), nfx (npx))
		allocate (nnz (npz), nfz (npz))


		call calc_size (model_s%nx, npx, nnx, nfx, .true.)
		call calc_size (model_s%nz, npz, nnz, nfz, .true.)

		call calc_size (model_s%nx, npx, n_nx, n_fx)
		call calc_size (model_s%nz, npz, n_nz, n_fz)

		!setup distrib
		nrec_found=0

		do ith = 1, npx*npz
			distrib_s(ith)%hx = -1
			distrib_s(ith)%hz = -1
			distrib_s(ith)%lx = -1
			distrib_s(ith)%lz = -1

			call helix2cart ((/npx, npz/), ith, i)

			distrib_s(ith)%fx = nfx(i(1))
			distrib_s(ith)%fz = nfz(i(2))
			distrib_s(ith)%nx = nnx(i(1))
			distrib_s(ith)%nz = nnz(i(2))

			distrib_s(ith)%f_x = n_fx(i(1))
			distrib_s(ith)%f_z = n_fz(i(2))
			distrib_s(ith)%n_x = n_nx(i(1))
			distrib_s(ith)%n_z = n_nz(i(2))

			distrib_s(ith)%ixsrc = model_s%ixsrc - distrib_s(ith)%fx
			distrib_s(ith)%izsrc = model_s%izsrc - distrib_s(ith)%fz

			allocate(distrib_s(ith)%dx(nnx(i(1))))
			allocate(distrib_s(ith)%dz(nnz(i(2))))

			distrib_s(ith)%dx = model_s%dx (nfx (i(1)) + 1 : nfx (i(1)) + nnx (i(1)))
			distrib_s(ith)%dz = model_s%dz (nfz (i(2)) + 1 : nfz (i(2)) + nnz (i(2)))
			distrib_s(ith)%dt = model_s%dt

			! see how many receivers are in the computational domain
			nullify(distrib_s(ith)%recs)
			nr=0
			do irec = 1, size (model_s%recs,2)

				if (model_s%recs (1,irec) >  distrib_s(ith)%f_x .and. &
					model_s%recs (1,irec) <= distrib_s(ith)%f_x + distrib_s(ith)%n_x .and. &
					model_s%recs (2,irec) >  distrib_s(ith)%f_z .and. &
					model_s%recs (2,irec) <= distrib_s(ith)%f_z + distrib_s(ith)%n_z ) then

					model_s%rec_ith (irec) = ith
					nr = nr+1
					model_s%rec_i (irec) = nr
				end if
			end do

			! map receiver positions to padded computational domain
			if(nr>0) then
				allocate(distrib_s(ith)%recs(2,nr))
				nr=0
				do irec=1,size(model_s%recs,2)
					if (model_s%recs(2,irec) > distrib_s(ith)%f_z .and. &
						model_s%recs(2,irec) <= distrib_s(ith)%f_z+distrib_s(ith)%n_z .and. &
						model_s%recs(1,irec) > distrib_s(ith)%f_x .and. &
						model_s%recs(1,irec) <= distrib_s(ith)%f_x+distrib_s(ith)%n_x ) then

							nr = nr+1	
							distrib_s (ith)%recs (1,nr) = model_s%recs (1,irec) - distrib_s (ith)%fx
							distrib_s (ith)%recs (2,nr) = model_s%recs (2,irec) - distrib_s (ith)%fz
							nrec_found = nrec_found+1
					end if
				end do
			end if

			if(i(1)>1) &
				call cart2helix ((/npx,npz/), (/i(1)-1, i(2)/), distrib_s (ith)%lx)
			if(i(2)>1) &
				call cart2helix ((/npx,npz/), (/i(1), i(2)-1/), distrib_s (ith)%lz)

			if(i(1)<npx) &
				call cart2helix ((/npx,npz/), (/i(1)+1, i(2)/), distrib_s (ith)%hx)
			if(i(2)<npz) &
				call cart2helix ((/npx,npz/), (/i(1), i(2)+1/), distrib_s (ith)%hz)

			bx = 1
			if (distrib_s(ith)%lx == -1) bx = 21
			bz = 1
			if (distrib_s(ith)%lz == -1) bz = 21

			ex = distrib_s(ith)%nx
			ez = distrib_s(ith)%nz
			if (distrib_s(ith)%hx == -1) ex = ex-20
			if (distrib_s(ith)%hz == -1) ez = ez-20
		end do

		deallocate (b_l_x, b_l_z, b_u_x, b_u_z)
		deallocate (n_nx, n_nz, n_fx, n_fz)
		deallocate (nnx, nnz, nfx, nfz)

		if (nrec_found /= size (model_s%recs, 2)) then
		 call seperr("Not all receivers in computational domain")
		end if


	end subroutine

	
	!calculate the size and offset for a region along a given axis
	subroutine calc_size (nlen, np, n, f, halo)
		integer :: i, f(:), n(:), np, nlen, b, nmin
		logical, optional :: halo

		b=0
		nmin=nlen/np
		do i=1,np
			n(i)=nmin
			if(mod(nlen,np) >= i) n(i)=n(i)+1
			f(i)=b 
			b=b+n(i)
		end do

		if(present(halo)) then
			!now add "halo"
			!left halo
			do i=2,np
				f(i)=f(i)-3
				n(i)=n(i)+3
			end do
		
			if(np>1) then
				!right side
				do i=1,np-1
					n(i)=n(i)+2
				end do
			end if
		end if
	end subroutine
end module
