module helutil_mod

	use owilson_mod
	use helicon
	implicit none

contains

! Factor 1-D 1st order FD scheme, using wilson-burg factorization.
function lapfac1d (eps, b, a1, n1, na)   result (aa)
	type (filter)                       :: aa, lap
	real,                   intent (in) :: eps, b, a1
	integer,                intent (in) :: na, n1
	integer                             :: i
	real                                :: a0, c0

	call allocatehelix (lap, 1)	!  FD filter
	c0         = b + eps		! zero lag coeff.
	lap%lag    = (/ 1 /)	! lag(1)= 1 # one side only
	lap%flt(1) = a1

	call allocatehelix (aa, na)	!  FD derivative
	aa%flt = 0.

	do i = 1, na  
		aa%lag (i     ) = i				! early lags (first row)
	end do 

	call owilson_init	(10 * n1)
	call owilson_factor	(20, c0, lap, a0, aa)
	call owilson_close	()
	call deallocatehelix(lap)
end function 

! Factor 2-D 2nd order FD scheme, using wilson-burg factorization.
function lapfac2 (eps, b, a1, a2, n1, na)   result (aa)
	type (filter)                       :: aa, lap
	real,                   intent (in) :: eps, b, a1, a2
	integer,                intent (in) :: n1, na
	integer                             :: i
	real                                :: a0, c0

	call allocatehelix (lap, 2)	!  FD filter
	c0         = b + eps		! zero lag coeff.
	lap%lag    = (/ 1, n1 /)	! lag(1)= 1; lag(2)=n1  # one side only
	lap%flt(1) = a1
	lap%flt(2) = a2

	call allocatehelix (aa, 2*na)	!  FD derivative
	aa%flt = 0.

	do i = 1, na  
		aa%lag (i     ) = i				! early lags (first column)
		aa%lag (na + i) = n1 + i - na	! late lags (second column)
	end do 

	call owilson_init	(10 * n1)
	call owilson_factor	(20, c0, lap, a0, aa)
	call owilson_close	()
	call deallocatehelix(lap)
end function 

! Factor 2-D  4th order FD scheme, using wilson-burg factorization.
function lapfac4 (eps, c, a1_1, a2_1, a1_2, a2_2, n1, na)   result (aa)
	type (filter)                       :: aa, lap
	real,                   intent (in) :: eps, c, a1_1, a2_1, a1_2, a2_2
	integer,                intent (in) :: n1, na
	integer                             :: i
	real                                :: a0, c0

	call allocatehelix (lap, 4)	!  FD filter
	c0         = c + eps		! zero lag coeff.
	lap%lag    = (/ 1, 2, n1, 2*n1 /)	! lag(1)= 1, lag(2)=2, lag(3)=n1, lag(4)=2*n1  # one side only
	lap%flt(1) = a1_1
	lap%flt(2) = a1_2
	lap%flt(3) = a2_1
	lap%flt(4) = a2_2

	call allocatehelix (aa, 4*na)	!  FD derivative
	aa%flt = 0.

	do i = 1, na  
		aa%lag (i     )   = i				! early lags (first column)
		aa%lag (na + i)   = n1 + i - na		! middle lags (second column)
		aa%lag (2*na + i) = n1 + i			! middle lags (second column)
		aa%lag (3*na + i) = 2*n1 + i - na	! late lags (third column)
	end do 

	call owilson_init	(10 * n1)
	call owilson_factor	(20, c0, lap, a0, aa, .false.)
	call owilson_close	()
	call deallocatehelix(lap)

end function 

! Factor 3-D  4th order FD scheme, using wilson-burg factorization.
function lapfac3d_4 (eps, c, a1_1, a2_1, a3_1, a1_2, a2_2, a3_2, n1, n2, na)   result (aa)
	type (filter)                       :: aa, lap
	real,                   intent (in) :: eps, c, a1_1, a2_1, a3_1, a1_2, a2_2, a3_2
	integer,                intent (in) :: n1, n2, na
	integer                             :: i
	real                                :: a0, c0

	call allocatehelix (lap, 6)	!  FD filter
	c0         = c + eps		! zero lag coeff.
	lap%lag    = (/ 1, 2, n1, 2*n1, n1*n2, 2*n1*n2 /)	! lag(1)= 1, lag(2)=2, lag(3)=n1, lag(4)=2*n1, lag(5)=n1*n2, lag(6)=2*n1*n2  # one side only
	lap%flt(1) = a1_1
	lap%flt(2) = a1_2
	lap%flt(3) = a2_1
	lap%flt(4) = a2_2
	lap%flt(5) = a3_1
	lap%flt(6) = a3_2

	call allocatehelix (aa, 24*na)	!  FD derivative
	aa%flt = 0.

	do i = 1, na  
		! first 2d panel
		aa%lag (0*na  + i) = i
		aa%lag (1*na  + i) = n1 + i - na
		aa%lag (2*na  + i) = n1 + i
		aa%lag (3*na  + i) = 2*n1 + i - na
		aa%lag (4*na  + i) = 2*n1 + i


		! second 2d panel
		aa%lag (5*na  + i) = n1*n2 - 3*n1 + i - na
		aa%lag (6*na  + i) = n1*n2 - 3*n1 + i
		aa%lag (7*na  + i) = n1*n2 - 2*n1 + i - na
		aa%lag (8*na  + i) = n1*n2 - 2*n1 + i
		aa%lag (9*na  + i) = n1*n2 - n1 + i - na
		aa%lag (10*na + i) = n1*n2 - n1 + i
		aa%lag (11*na + i) = n1*n2 + i - na
		aa%lag (12*na + i) = n1*n2 + i
		aa%lag (13*na + i) = n1*n2 + n1 + i - na
		aa%lag (14*na + i) = n1*n2 + n1 + i
		aa%lag (15*na + i) = n1*n2 + 2*n1 + i - na
		aa%lag (16*na + i) = n1*n2 + 2*n1 + i

		! third 2d panel
		aa%lag (17*na + i) = 2*n1*n2 - 3*n1 + i - na
		aa%lag (18*na + i) = 2*n1*n2 - 3*n1 + i
		aa%lag (19*na + i) = 2*n1*n2 - 2*n1 + i - na
		aa%lag (20*na + i) = 2*n1*n2 - 2*n1 + i
		aa%lag (21*na + i) = 2*n1*n2 - n1 + i - na
		aa%lag (22*na + i) = 2*n1*n2 - n1 + i
		aa%lag (23*na + i) = 2*n1*n2 + i - na
	end do 

	call owilson_init	(10 * n1 * n2)
	call owilson_factor	(20, c0, lap, a0, aa, .false.)
	call owilson_close	()
	call deallocatehelix(lap)

end function 


! convolution and correlation function for floats
	subroutine convolve (f, g, h, n, m)
		integer, intent(in)					:: n, m
		real, dimension(:), intent(in)	:: f, g
		real, dimension(:), intent(out)	:: h
		integer 							:: i, j

		h = 0.
	
		do i = 0, n-1
			do j = 0, m-1
				h (i+j+1) = h (i+j+1) + f(i+1) * g(j+1)
			end do
		end do
	end subroutine

	subroutine correlate (f, g, h, n)
		integer, intent(in)					:: n
		real, dimension(:), intent(in)	:: f, g
		real, dimension(:), intent(out)	:: h
		integer								:: i, j
	
		h = 0.
	
		do i = 0, n-1
			do j = i, n-1
				h (j-i+1) = h (j-i+1) + f(i+1) * g(j+1)
			end do
		end do
	end subroutine

	function getvmax3d (f) result (vmax)
		real, dimension(:,:,:), intent (in)	:: f
		real	:: vmax
		integer	:: i, j, k

		vmax = 0.

		do k = 1, size (f, 3)
			do j = 1, size (f, 2)
				do i = 1, size (f, 1)
				if (f (i, j, k) .gt. vmax) vmax = f (i, j, k)
				end do
			end do
		end do

	end function

end module helutil_mod
